"""
Implementation of buffered end-to-end filter.
"""
from typing import Tuple

import torch

from motrack_motion.filter.base import State
from motrack_motion.filter.buffer import ObservationBuffer, BufferedStateModel
from motrack_motion.models.architectures.transfilter import TransFilter, pad_sequence


class BufferedTransFilter(BufferedStateModel):
    """
    Implementation of buffered TransFilter.
    """
    def _create_uncertainty_tensor(self, x: torch.Tensor) -> torch.Tensor:
        return self._uncertainty_multiplier * torch.ones_like(x).to(x)

    def initiate(self, measurement: torch.Tensor) -> State:
        buffer = self._create_buffer(measurement)
        return buffer, None, None, None, None

    @torch.no_grad()
    def multistep_predict(self, state: State, n_steps: int) -> State:
        assert n_steps == 1, 'Multistep prediction is not supported!'

        buffer, _, _, zs, _ = state
        buffer: ObservationBuffer

        if not buffer.has_input:
            raise BufferError('Buffer does not have an input!')

        x_obs, ts_obs, ts_unobs = buffer.get_input(n_steps)
        if ts_obs.shape[0] == 1:
            # Only one bbox in history - using baseline (propagate last bbox) instead of NN model
            x_unobs_mean_hat, x_unobs_var_hat = self._baseline(x_obs, ts_obs, ts_unobs, multistep=True)
            return buffer, x_unobs_mean_hat, x_unobs_var_hat, None, None

        if self._recursive_inverse:
            t_obs_last_int, t_unobs_last_int = int(ts_obs[-1, 0, 0]), int(ts_unobs[-1, 0, 0])
            ts_unobs = torch.tensor(list(range(t_obs_last_int + 1, t_unobs_last_int + 1)), dtype=torch.float32).view(-1, 1, 1)

        # Preprocess
        t_x_obs, _, t_ts_obs, t_ts_unobs, *_ = self._transform.apply(data=[x_obs, None, ts_obs, ts_unobs, None], shallow=False)
        t_x_obs, t_ts_obs, mask = pad_sequence(t_x_obs, t_ts_obs)

        # Inference
        self._model: TransFilter
        t_x_obs, t_ts_obs, t_ts_unobs, mask = \
            t_x_obs.to(self._accelerator), t_ts_obs.to(self._accelerator), t_ts_unobs.to(self._accelerator), mask.to(self._accelerator)
        zs, t_obs_last = self._model.encode_trajectory(t_x_obs, t_ts_obs, mask=mask)
        t_x_prior_mean = self._model.predict(zs, t_obs_last, t_ts_unobs)
        t_x_prior_mean = t_x_prior_mean.detach().cpu()

        # Postprocess
        t_x_prior_var = self._create_uncertainty_tensor(t_x_prior_mean)
        _, prior_mean, *_ = self._transform.inverse(data=[x_obs, t_x_prior_mean, None], shallow=False)
        prior_var = self._transform.inverse_var(t_x_prior_var, additional_data=[x_obs, None], shallow=False)
        prior_mean = prior_mean[-1]  # Removing temporal dimension after inverse transform
        prior_var = prior_var[-1]

        return buffer, prior_mean, prior_var, zs, t_x_prior_mean[-1]

    def predict(self, state: State) -> State:
        buffer, prior_mean, prior_var, zs, t_prior_mean = self.multistep_predict(state, n_steps=1)
        return buffer, prior_mean[0], prior_var[0], zs, (t_prior_mean[0] if t_prior_mean is not None else None)

    def singlestep_to_multistep_state(self, state: State) -> State:
        buffer, prior_mean, prior_var, zs, t_prior_mean  = state
        return buffer, prior_mean.unsqueeze(0), prior_var.unsqueeze(0), zs, t_prior_mean.unsqueeze(0)

    @torch.no_grad()
    def update(self, state: State, measurement: torch.Tensor) -> State:
        buffer, _, _, zs, t_x_prior_mean = state
        buffer: ObservationBuffer

        x_obs, ts_obs, ts_unobs = buffer.get_input(1)
        buffer.push(measurement)

        if ts_obs.shape[0] == 1:
            # Only one bbox in history - using baseline (propagate last bbox) instead of NN model
            x_unobs_mean_hat, x_unobs_var_hat = self._baseline(x_obs, ts_obs, ts_unobs)
            return buffer, x_unobs_mean_hat, x_unobs_var_hat, None, None
        assert zs is not None

        # Preprocess
        _, t_measurement, t_ts_obs, t_ts_unobs, _ = self._transform.apply(data=[x_obs, measurement.view(1, 1, -1), ts_obs, ts_unobs, None],
                                                                       shallow=False)
        t_obs_last = t_ts_obs[-1, 0, 0]
        t_x_prior_mean = t_x_prior_mean.view(1, 1, -1)

        # Inference
        t_measurement, zs, t_obs_last, t_x_prior_mean, t_ts_unobs = \
            t_measurement.to(self._accelerator), zs.to(self._accelerator), t_obs_last.to(self._accelerator), t_x_prior_mean.to(self._accelerator), \
            t_ts_unobs.to(self._accelerator)
        t_posterior_mean = self._model.update(zs, t_obs_last, t_x_prior_mean, t_measurement, t_ts_unobs)
        t_posterior_mean, t_x_prior_mean = t_posterior_mean.detach().cpu(), t_x_prior_mean.detach().cpu()

        # Postprocess
        t_posterior_var = self._create_uncertainty_tensor(t_posterior_mean)
        _, posterior_mean, *_ = self._transform.inverse(data=[x_obs, t_posterior_mean.view(1, 1, -1), None], shallow=False)
        posterior_var = self._transform.inverse_var(t_posterior_var, additional_data=[posterior_mean, None], shallow=False)

        posterior_mean = posterior_mean[-1, 0, :]
        posterior_var = posterior_var[-1, 0, :]

        return buffer, posterior_mean, posterior_var, zs, None

    def missing(self, state: State) -> State:
        buffer, mean, var, z_state, _ = state
        buffer.increment()
        return buffer, mean, var, z_state, None

    def project(self, state: State) -> Tuple[torch.Tensor, torch.Tensor]:
        _, mean, var, _, _ = state
        return mean, var

    def _baseline(
        self,
        x_obs: torch.Tensor,
        ts_obs: torch.Tensor,
        ts_unobs: torch.Tensor,
        multistep: bool = False
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        """
        In case there are not enough measurements (e.g. only one measurement is present),
        `baseline` is used instead of a NN model. Baseline just copies last observation.

        Args:
            x_obs: Observations
            ts_obs: Observations' time points
            ts_unobs: Unobserved (prediction) time points
            multistep: Multistep

        Returns:
            - Mean prediction
            - Standard deviation prediction
            - latent state
        """
        x_unobs_mean_hat = torch.stack([x_obs[-1].clone() for _ in range(ts_unobs.shape[0])]).to(ts_obs)
        x_unobs_var_hat = self._create_uncertainty_tensor(x_unobs_mean_hat)
        x_unobs_mean_hat, x_unobs_var_hat = x_unobs_mean_hat[:, 0, :], x_unobs_var_hat[:, 0, :]  # Remove batch
        if not multistep:
            x_unobs_mean_hat, x_unobs_var_hat = x_unobs_mean_hat[0], x_unobs_var_hat[0]

        return x_unobs_mean_hat, x_unobs_var_hat


def run_test() -> None:
    # pylint: disable=import-outside-toplevel
    from motrack_motion.datasets.transforms.base import IdentityTransform

    smf = BufferedTransFilter(
        model=TransFilter(
            input_dim=4,
            d_model=16,
            output_dim=4,
            t_scale=5.0
        ),
        buffer_size=5,
        accelerator='cpu',
        transform=IdentityTransform()
    )

    measurements = torch.randn(5, 4, dtype=torch.float32)

    # Initiate test
    state = smf.initiate(measurements[0])

    for i in range(1, 5):
        # Predict test
        state = smf.predict(state)
        mean_hat, cov_hat = smf.project(state)
        assert mean_hat.shape == (4,) and cov_hat.shape == (4,)

        # Update test
        state = smf.update(state, measurements[i])
        mean_updated, cov_updated = smf.project(state)
        assert mean_updated.shape == (4,) and cov_updated.shape == (4,)

    for i in range(5, 10):
        # Predict test
        state = smf.predict(state)
        mean_hat, cov_hat = smf.project(state)
        assert mean_hat.shape == (4,) and cov_hat.shape == (4,)

        # Missing test
        state = smf.missing(state)
        mean_updated, cov_updated = smf.project(state)
        assert mean_updated.shape == (4,) and cov_updated.shape == (4,)


if __name__ == '__main__':
    run_test()



"""
Transform factory method.
"""
from motrack_motion.datasets.transforms import (
    InvertibleTransform,
    InvertibleTransformWithVariance,
    IdentityTransform,
    BboxFirstOrderDifferenceTransform,
    BBoxStandardizationTransform,
    BBoxStandardizedFirstOrderDifferenceTransform,
    BBoxRelativeToLastObsTransform,
    BBoxStandardizedRelativeToLastObsTransform,
    BBoxCompositeTransform,
    BBoxJackOfAllTradesTransform,
)
from typing import Union


def transform_factory(name: str, params: dict) -> Union[InvertibleTransform, InvertibleTransformWithVariance]:
    """
    Create transform object based on given name and constructor parameters.

    Args:
        name: Transform name
        params: Transform parameters

    Returns:
        Transform object
    """
    catalog = {
        'identity': IdentityTransform,
        'composite': BBoxCompositeTransform,
        'first_difference': BboxFirstOrderDifferenceTransform,
        'standardization': BBoxStandardizationTransform,
        'standardized_first_difference': BBoxStandardizedFirstOrderDifferenceTransform,
        'relative_to_last_obs': BBoxRelativeToLastObsTransform,
        'standardized_relative_to_last_obs': BBoxStandardizedRelativeToLastObsTransform,
        'jack_of_all_trades': BBoxJackOfAllTradesTransform
    }

    cls = catalog[name]

    if name == 'composite':
        params['transforms'] = [transform_factory(child['name'], child['params']) for child in params['transforms']]

    return cls(**params)



"""
Building block: MultiLayerPerceptron
"""
from typing import Optional, List

import torch
from torch import nn


class MLP(nn.Module):
    """
    MultiLayerPerceptron where each linear layer contains:
    - Linear layer
    - LayerNorm
    - SiLU
    """
    def __init__(
        self,
        input_dim: int,
        hidden_dim: Optional[int] = None,
        output_dim: Optional[int] = None,
        n_layers: int = 1,
        activation: Optional[nn.Module] = nn.SiLU,
        residual: bool = False
    ):
        """
        Args:
            input_dim: Model Input dimension
            hidden_dim: Model hidden dimension
                - tensor dimension between layers
                - equal to `input_dim` if not defined
                - not relevant in case `n_layers` is equal to `1`
            output_dim: Model output dimension
                - equal to `input_dim` if not defined
            n_layers: Number of Perceptron layers
            residual: Use skip connections
        """
        super().__init__()
        self._residual = residual

        assert n_layers >= 1, f'Minimum number of layers is 1 but found {n_layers}'
        if output_dim is None:
            output_dim = input_dim
        if hidden_dim is None:
            hidden_dim = output_dim

        assert not residual or (input_dim == hidden_dim == output_dim)

        layers_args = [[hidden_dim, hidden_dim] for _ in range(n_layers)]
        layers_args[0][0] = input_dim
        layers_args[-1][1] = output_dim
        self._model = nn.Sequential(*[self._create_mlp_layer(*args, activation=activation) for args in layers_args])

    @staticmethod
    def _create_mlp_layer(input_dim: int, output_dim: int, activation: Optional[nn.Module]) -> nn.Module:
        """
        Constructs previously defined MLP layer

        Args:
            input_dim: Input dimension
            output_dim: Output dimension
            activation: Activation function

        Returns:
            Perceptron
        """
        layers: List[nn.Module] = [
            nn.Linear(input_dim, output_dim),
            nn.LayerNorm(output_dim)
        ]


        if activation is not None:
            layers.append(activation())

        return nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        if not self._residual:
            return self._model(x)

        return x + self._model(x)


def main() -> None:
    mlp = MLP(
        input_dim=5,
        hidden_dim=10,
        n_layers=2
    )

    x = torch.randn(32, 5)  # (batch_size, dim)
    expected_shape = (32, 5)

    output = mlp(x)
    assert output.shape == expected_shape, f'Expected shape {expected_shape} but found {output.shape}!'


if __name__ == '__main__':
    main()



"""
Filter factory.
"""
from torch import nn

from motrack_motion.datasets import transforms
from motrack_motion.filter.end_to_end_filter import BufferedE2EFilter
from motrack_motion.filter.transfilter import BufferedTransFilter
from motrack_motion.filter.base import StateModelFilter

FILTER_CATALOG = {
    'end_to_end': BufferedE2EFilter,
    'transfilter': BufferedTransFilter
}


def filter_factory(
    name: str,
    params: dict,
    model: nn.Module,
    transform: transforms.InvertibleTransformWithVariance,
) -> StateModelFilter:
    """
    Creates filter by given name.

    Args:
        name: Dataset name
        params: Filter parameters
        model: Filter core model
        transform: Filter preprocess - postprocess

    Returns:
        Initialized filter
    """
    name = name.lower()

    cls = FILTER_CATALOG[name]
    return cls(
        **params,
        model=model,
        transform=transform
    )



"""
Implementation of the TransFilter architecture.
"""
import math
from typing import Optional, Union, Dict, Tuple, List

import torch
from torch import nn

from motrack_motion.library.building_blocks.attention import TemporalFirstMultiHeadCrossAttention
from motrack_motion.library.building_blocks.mlp import MLP


class ReversedPositionalEncoding(nn.Module):
    """
    Reversed positional encoding. Same as the standard positional encoding
    make sure the last element always gets the same encoding.

    Note: Missing values should be padded before usage.
    """
    def __init__(self, d_model: int, dropout: float = 0.1, max_len: int = 5000):
        """
        Args:
            d_model: Model size
            dropout: Dropout rate
            max_len: Maximum sequence length
        """
        super().__init__()
        self.dropout = nn.Dropout(p=dropout)

        # Generate positions in reverse order
        position = torch.arange(max_len - 1, -1, -1).unsqueeze(1)
        div_term = torch.exp(torch.arange(0, d_model, 2) * (-math.log(10000.0) / d_model))
        pe = torch.zeros(max_len, 1, d_model)
        pe[:, 0, 0::2] = torch.sin(position * div_term)
        pe[:, 0, 1::2] = torch.cos(position * div_term)
        self.register_buffer('pe', pe)

    @property
    def encoding(self) -> torch.Tensor:
        """
        Get the positional encoding vector.

        Returns:
            Positional encoding vector
        """
        return self.pe

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = x + self.pe[-x.size(0):]
        return self.dropout(x)


def pad_sequence(x: torch.Tensor, t: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
    """
    Pads the input sequence based on the observation time points.

    Args:
        x: Observation features
        t: Observation time points

    Returns:
        - Padded observation features (with zeros)
        - Interpolated time points
        - Mask for the padded observation features
    """
    batch_size = t.shape[1]
    t_first, t_last = int(t[0, 0, 0]), int(t[-1, 0, 0])
    length = (t_last - t_first) + 1
    t_padded = torch.tensor(list(range(t_first, t_first + length)), dtype=torch.float32) \
        .view(-1, 1, 1) \
        .repeat(1, batch_size, 1)

    x_padded = torch.zeros(length, batch_size, x.shape[-1])
    indices = [int(t[i, 0, 0]) - t_first for i in range(t.shape[0])]
    for i_orig, i_pad in enumerate(indices):
        x_padded[i_pad] = x[i_orig]

    padded_indices = [i for i in range(length) if i not in indices]
    mask = nn.Transformer.generate_square_subsequent_mask(length)
    mask[:, padded_indices] = -torch.inf

    x_padded, t_padded, mask = x_padded.to(x), t_padded.to(t), mask.to(x)
    return x_padded, t_padded, mask


class TransFilterUpdateDecoder(nn.Module):
    """
    TransFilter decoder. Performs correction on the prior predictions after the observations are observed.
    """
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        n_heads: int,
        dropout: float = 0.1,
        n_layers: int = 2
    ):
        super().__init__()

        self._mhca = TemporalFirstMultiHeadCrossAttention(n_heads=n_heads)
        self._norm = nn.LayerNorm(input_dim)
        self._activation = nn.SiLU()
        self._dropout = nn.Dropout(p=dropout)
        self._head = nn.Sequential(
            MLP(input_dim=input_dim, output_dim=input_dim, n_layers=n_layers - 1),
            nn.Linear(input_dim, output_dim)
        )

    def forward(self, x_target: torch.Tensor, x_input: torch.Tensor) -> torch.Tensor:
        # MHCA
        x = self._mhca(q=x_target, k=x_input, v=x_input)
        x = self._norm(x)
        x = self._activation(x)
        x = self._dropout(x)
        x = x_target + x

        # FFN
        return self._head(x)


class TransFilter(nn.Module):
    """
    TransFilter model architecture.
    """
    def __init__(
        self,
        input_dim: int,
        output_dim: int,

        # Model (global)
        d_model: int,
        activation: Union[nn.Module, str] = nn.SiLU,
        t_scale: float = 30.0,
        pooling: str = 'mean',

        # Positional encoding
        pe_dropout: float = 0.1,
        pe_max_len: int = 120,

        # Encoder
        enc_n_layers: int = 6,
        enc_n_heads: int = 8,
        enc_d_model: Optional[int] = None,
        enc_dropout: float = 0.1,

        # Prediction head
        pred_steps: int = 30,
        pred_n_layers: int = 2,

        # Decoder
        dec_n_heads: int = 8,
        dec_dropout: float = 0.1,
        dec_n_layers: int = 2
    ):
        super().__init__()

        # Defaults
        enc_d_model = d_model if enc_d_model is None else enc_d_model
        activation = activation() if not isinstance(activation, str) else activation

        self._t_scale = t_scale
        self._output_dim = output_dim
        self._steps = pred_steps

        # Architecture
        assert pooling in ['mean', 'max', 'last'], f'Invalid pooling option "{pooling}"!'
        self._pooling = pooling
        self._input_embedding = MLP(
            input_dim=input_dim + 1,  # includes time
            hidden_dim=d_model,
            output_dim=d_model
        )
        self._pos_encoding = ReversedPositionalEncoding(
            d_model=d_model,
            dropout=pe_dropout,
            max_len=pe_max_len
        )
        self._encoder = nn.TransformerEncoder(
            encoder_layer=nn.TransformerEncoderLayer(
                d_model=d_model,
                nhead=enc_n_heads,
                dim_feedforward=enc_d_model,
                dropout=enc_dropout,
                activation=activation
            ),
            num_layers=enc_n_layers
        )

        self._predict_head = nn.Sequential(
            MLP(input_dim=d_model, output_dim=d_model, n_layers=pred_n_layers - 1),
            nn.Linear(d_model, pred_steps * output_dim)
        )

        self._target_embedding = MLP(
            input_dim=2 * output_dim + 1,  # Innovation (4) + time (1)
            hidden_dim=d_model,
            output_dim=d_model
        )
        self._decoder = TransFilterUpdateDecoder(
            input_dim=d_model,
            output_dim=output_dim,
            n_heads=dec_n_heads,
            dropout=dec_dropout,
            n_layers=dec_n_layers
        )

    def encode_trajectory(self, x_obs: torch.Tensor, t_obs: torch.Tensor, mask: Optional[torch.Tensor] = None) \
            -> Tuple[torch.Tensor, torch.Tensor]:
        t_obs_last = t_obs[-1, 0, 0]

        with torch.no_grad():
            t_obs = (t_obs - t_obs_last) / self._t_scale

        xt_input = torch.cat([x_obs, t_obs], dim=-1)
        z_input = self._input_embedding(xt_input)
        z_input = z_input + self._pos_encoding(z_input)

        if mask is None:
            # Create causal mask
            mask = nn.Transformer.generate_square_subsequent_mask(z_input.shape[0]).to(z_input)

        zs = self._encoder(z_input, mask)
        return zs, t_obs_last

    def _pool(self, zs: torch.Tensor) -> torch.Tensor:
        if self._pooling == 'mean':
            return torch.mean(zs, dim=0)
        if self._pooling == 'max':
            return torch.max(zs, dim=0)[0]
        if self._pooling == 'last':
            return zs[-1, :, :]

        raise AssertionError('Invalid Program State!')

    def get_prediction_indices(self, t_obs_last: torch.Tensor, t_unobs: torch.Tensor) -> List[int]:
        t_unobs_first = int(t_obs_last) + 1
        return [min(self._steps - 1, int(t) - t_unobs_first) for t in t_unobs[:, 0, 0].detach()]

    def predict(self, zs: torch.Tensor, t_obs_last: torch.Tensor, t_unobs: torch.Tensor) -> torch.Tensor:
        z_agg = self._pool(zs)
        x_prior = self._predict_head(z_agg) # (B, steps * output_dim)
        x_prior = x_prior.view(-1, self._steps, self._output_dim) # (B, steps, output_dim)
        x_prior = torch.transpose(x_prior, 0, 1)

        if t_unobs.shape[0] < self._steps:
            indices = self.get_prediction_indices(t_obs_last, t_unobs)
            x_prior = x_prior[indices]
        return x_prior

    def update(self, zs: torch.Tensor, t_obs_last: torch.Tensor, x_prior: torch.Tensor, x_unobs: torch.Tensor, t_unobs: torch.Tensor) -> torch.Tensor:
        with torch.no_grad():
            innovation = x_unobs - x_prior
            t_unobs = (t_unobs - t_obs_last) / self._t_scale

        # Update step
        xti_target = torch.cat([x_unobs, t_unobs, innovation], dim=-1)
        z_target = self._target_embedding(xti_target)
        return self._decoder(z_target, zs)

    def forward(self, x_obs: torch.Tensor, t_obs: torch.Tensor, x_unobs: torch.Tensor, t_unobs: torch.Tensor, mask: Optional[torch.Tensor] = None) \
            -> Dict[str, torch.Tensor]:
        assert torch.abs(t_obs[1:, 0, 0] - t_obs[:-1, 0, 0] - 1).sum() < 1e-6, f'Sequence is not padded! {t_obs[:, 0, 0]=}'

        zs, t_obs_last = self.encode_trajectory(x_obs, t_obs, mask=mask)
        x_prior = self.predict(zs, t_obs_last, t_unobs)
        x_posterior = self.update(zs, t_obs_last, x_prior, x_unobs, t_unobs)

        return {
            'z_traj': zs,
            'x_prior': x_prior,
            'x_posterior': x_posterior
        }


def run_test() -> None:
    # Test model
    model = TransFilter(
        input_dim=4,
        output_dim=4,
        d_model=8
    )

    x_obs = torch.randn(10, 2, 4)
    t_obs = torch.tensor(list(range(1, 11)), dtype=torch.float32).view(-1, 1, 1).repeat(1, 2, 1)
    x_unobs = torch.randn(5, 2, 4)
    t_unobs = torch.tensor(list(range(16, 21)), dtype=torch.float32).view(-1, 1, 1).repeat(1, 2, 1)

    outputs = model(x_obs, t_obs, x_unobs, t_unobs)
    z_traj = outputs['z_traj']
    x_prior = outputs['x_prior']
    x_posterior = outputs['x_posterior']

    assert z_traj.shape == (10, 2, 8), f'{z_traj.shape=}'
    assert x_prior.shape == (5, 2, 4), f'{x_prior.shape=}'
    assert x_posterior.shape == (5, 2, 4), f'{x_posterior.shape=}'

    # Test padding
    x = torch.randn(3, 2, 4)
    t = torch.tensor([3, 4, 7], dtype=torch.float32).view(-1, 1, 1).repeat(1, 2, 1)
    t_expected = torch.tensor([3, 4, 5, 6, 7], dtype=torch.float32).view(-1, 1, 1).repeat(1, 2, 1)
    x_padded, t_padded, mask = pad_sequence(x, t)

    assert x_padded.shape == (5, 2, 4), f'{x.shape=}'
    assert t_padded.shape == (5, 2, 1), f'{t.shape}'
    assert torch.abs(t_padded - t_expected).sum() < 1e-6, f'{t=}'

    outputs = model(x_padded, t_padded, x_unobs, t_unobs, mask=mask)
    z_traj = outputs['z_traj']
    x_prior = outputs['x_prior']
    x_posterior = outputs['x_posterior']

    assert z_traj.shape == (5, 2, 8), f'{z_traj.shape=}'
    assert x_prior.shape == (5, 2, 4), f'{x_prior.shape=}'
    assert x_posterior.shape == (5, 2, 4), f'{x_posterior.shape=}'


if __name__ == '__main__':
    run_test()



"""
Motrack-motion Lightning module training utility.
"""
from dataclasses import dataclass, field
from typing import Tuple, Union, Dict, Optional

import numpy as np
import pytorch_lightning as pl
import torch

from motrack_motion.models.train.meter import MetricMeter
from motrack_motion.utils import torch_helper


def extract_mean_and_var(bboxes_unobs_hat: torch.Tensor, bounded_variance: bool = False, bounded_value: float = 0.01) \
        -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Helper function for Gaussian model postprocess

    Args:
        bboxes_unobs_hat: Prediction
        bounded_variance: Bound minimal variance
        bounded_value: Bounded variance value

    Returns:
        bboxes_hat_mean, bboxes_hat_var
    """
    bboxes_unobs_hat = bboxes_unobs_hat.view(*bboxes_unobs_hat.shape[:-1], -1, 2)
    bboxes_unobs_hat_mean = bboxes_unobs_hat[..., 0]
    bboxes_unobs_hat_log_var = bboxes_unobs_hat[..., 1]

    if not bounded_variance:
        bboxes_unobs_hat_var = torch.exp(bboxes_unobs_hat_log_var)
    else:
        bboxes_unobs_hat_var = bounded_value + (1 - bounded_value) * torch.nn.functional.softplus(bboxes_unobs_hat_log_var)

    return bboxes_unobs_hat_mean, bboxes_unobs_hat_var


def extract_mean_and_std(bboxes_unobs_hat: torch.Tensor, bounded_variance: bool = False, bounded_value: float = 0.01) \
        -> Tuple[torch.Tensor, torch.Tensor]:
    """
    Helper function for Gaussian model postprocess

    Args:
        bboxes_unobs_hat: Prediction
        bounded_variance: Bound minimal variance
        bounded_value: Bounded variance value

    Returns:
        bboxes_hat_mean, bboxes_hat_std
    """
    bboxes_unobs_hat_mean, bboxes_unobs_hat_var = extract_mean_and_var(bboxes_unobs_hat, bounded_variance, bounded_value)
    bboxes_unobs_hat_std = torch.sqrt(bboxes_unobs_hat_var)
    return bboxes_unobs_hat_mean, bboxes_unobs_hat_std


@dataclass
class LightningTrainConfig:
    """
    Training configuration for LightningForecaster.
    """
    loss_name: str = field(default='mse')
    loss_params: dict = field(default_factory=dict)

    learning_rate: float = field(default=1e-3)
    sched_lr_gamma: float = field(default=1.0)
    sched_lr_step: int = field(default=1)
    n_warmup_epochs: int = field(default=0)

    optim_name: str = field(default='default')
    optim_additional_params: dict = field(default_factory=dict)

    weight_decay: float = field(default=0.0)

    n_train_steps: int = field(default=1000)


class LightningModuleBase(pl.LightningModule):
    """
    PytorchLightning module wrapper with some simple default utilities.
    """
    def __init__(self, train_config: LightningTrainConfig):
        """
        Args:
            train_config: Universal training config
        """
        super().__init__()
        self._train_config = train_config
        self._meter = MetricMeter()

    @property
    def n_params(self) -> int:
        """
        Gets number of model parameters.

        Returns:
            Return number of model parameters
        """
        trainable_parameters = filter(lambda p: p.requires_grad, self.parameters())
        return sum(np.prod(p.size()) for p in trainable_parameters)

    def on_validation_epoch_end(self) -> None:
        for name, value in self._meter.get_all():
            show_on_prog_bar = name.endswith('/loss')
            self.log(name, value, prog_bar=show_on_prog_bar)

    def configure_optimizers(self):
        optim_name = self._train_config.optim_name.lower()
        optim_catalog = {
            'default': torch.optim.Adam,
            'sgd': torch.optim.SGD,
            'adam': torch.optim.Adam,
            'adamw': torch.optim.AdamW
        }
        optim_cls = optim_catalog[optim_name]

        optimizer = optim_cls(
            params=self._model.parameters(),
            lr=self._train_config.learning_rate,
            weight_decay=self._train_config.weight_decay
        )

        def warmup_scheduler_func(current_step: int):
            """
            Warmup scheduler learning rate multiplier for warm-up steps.

            Args:
                current_step: Current step

            Returns:
                Learning rate multiplier.
            """
            if self._train_config.n_warmup_epochs == 0:
                return 1.0
            return current_step / (self._train_config.n_warmup_epochs * self._train_config.n_train_steps)

        scheduler_obj = torch.optim.lr_scheduler.SequentialLR(
           optimizer=optimizer,
           schedulers=[
               torch.optim.lr_scheduler.LambdaLR(
                   optimizer=optimizer,
                   lr_lambda=warmup_scheduler_func
               ),
               torch.optim.lr_scheduler.StepLR(
                   optimizer=optimizer,
                   step_size=self._train_config.sched_lr_step * self._train_config.n_train_steps,
                   gamma=self._train_config.sched_lr_gamma
               )
           ],
           milestones=[self._train_config.n_warmup_epochs * self._train_config.n_train_steps]
        )

        scheduler = {
            'scheduler': scheduler_obj,
            'interval': 'step',
            'frequency': 1
        }

        return [optimizer], [scheduler]

    def _log_lr(self) -> None:
        """
        Logs learning rate at the current step.
        """
        # noinspection PyTypeChecker
        optimizer: torch.optim.Optimizer = self.optimizers()
        if isinstance(optimizer, list):
            optimizer = optimizer[0]
        lr = torch_helper.get_optim_lr(optimizer)
        self.log('general/lr', lr)

    def _log_loss(self, loss: Union[torch.Tensor, Dict[str, torch.Tensor]], prefix: str, log_step: bool = True) -> None:
        """
        Helper function to log loss. Options:
        - Single value: logged as "{prefix}/loss"
        - Dictionary: for each key log value as "{prefix}/{key}"

        Args:
            loss: Loss
            prefix: Prefix (train or val)
        """
        assert prefix in ['training', 'val'], f'Invalid prefix value "{prefix}"!'

        if isinstance(loss, dict):
            assert 'loss' in loss, \
                f'When returning loss as dictionary it has to have key "loss". Found: {list(loss.keys())}'
            for name, value in loss.items():
                value = value.detach().cpu()
                assert not torch.isnan(value).any(), f'Got nan value for key "{name}"!'
                self._meter.push(f'{prefix}-epoch/{name}', value)
                if log_step:
                    self.log(f'{prefix}/{name}', value, prog_bar=False)
        else:
            loss = loss.detach().cpu()
            assert not torch.isnan(loss).any(), 'Got nan value!'
            loss = loss.detach().cpu()
            self._meter.push(f'{prefix}-epoch/loss', loss)
            if log_step:
                self.log(f'{prefix}/loss', loss, prog_bar=False)

    def _log_metrics(self, metrics: Optional[Dict[str, float]], prefix: str) -> None:
        """
        Helper function to log metrics. Input format:
        - Dictionary: for each key log value as "{prefix}-epoch/{key}"

        Args:
            prefix: Prefix (train or val)
        """
        if metrics is None:
            return

        assert prefix in ['training', 'val'], f'Invalid prefix value "{prefix}"!'
        for name, value in metrics.items():
            self._meter.push(f'{prefix}-metrics/{name}', value)



"""
Set of data transformations
"""
from motrack_motion.datasets.transforms.base import (
    Transform,
    InvertibleTransform,
    InvertibleTransformWithVariance,
    IdentityTransform
)
from motrack_motion.datasets.transforms.bbox import (
    BboxFirstOrderDifferenceTransform,
    BBoxStandardizationTransform,
    BBoxStandardizedFirstOrderDifferenceTransform,
    BBoxRelativeToLastObsTransform,
    BBoxStandardizedRelativeToLastObsTransform,
    BBoxCompositeTransform,
    BBoxJackOfAllTradesTransform
)
from motrack_motion.datasets.transforms.factory import transform_factory



"""
Implementation of buffered end-to-end filter.
"""
from typing import Tuple

import torch

from motrack_motion.filter.base import State
from motrack_motion.filter.buffer import ObservationBuffer, BufferedStateModel


class BufferedE2EFilter(BufferedStateModel):
    """
    Implementation of buffered end-to-end
    """

    def _get_ts(self) -> Tuple[torch.Tensor, torch.Tensor]:
        t_obs = torch.tensor([0], dtype=self._dtype).view(1, 1)
        t_unobs = torch.tensor([1], dtype=self._dtype).view(1, 1)
        return t_obs, t_unobs

    def initiate(self, measurement: torch.Tensor) -> State:
        buffer = self._create_buffer(measurement)
        return buffer, None, None, None, (None, None)

    @torch.no_grad()
    def multistep_predict(self, state: State, n_steps: int) -> State:
        assert n_steps == 1, 'Multistep prediction is not supported!'

        buffer, _, _, _, (t_cache, z_cache) = state
        buffer: ObservationBuffer

        if not buffer.has_input:
            raise BufferError('Buffer does not have an input!')

        x_obs, ts_obs, ts_unobs = buffer.get_input(n_steps)
        if ts_obs.shape[0] == 1:
            # Only one bbox in history - using baseline (propagate last bbox) instead of NN model
            x_unobs_mean_hat, x_unobs_std_hat, z1_prior = self._baseline(x_obs, ts_obs, ts_unobs, multistep=True)
            return buffer, x_unobs_mean_hat, x_unobs_std_hat, z1_prior, (t_cache, z_cache)

        use_cache = t_cache is not None and buffer.last_obs_time == t_cache
        ts_first = ts_obs[0, 0, 0]
        ts_obs = ts_obs - ts_first + 1
        ts_unobs = ts_unobs - ts_first + 1
        t_obs_last, t_unobs_last = int(ts_obs[-1, 0, 0]), int(ts_unobs[-1, 0, 0])
        if self._recursive_inverse:
            ts_unobs = torch.tensor(list(range(t_obs_last + 1, t_unobs_last + 1)), dtype=torch.float32).view(-1, 1, 1)

        t_x_obs, _, t_ts_obs, t_ts_unobs, *_ = self._transform.apply(data=[x_obs, None, ts_obs, ts_unobs, None], shallow=False)
        t_x_obs, t_ts_obs, t_ts_unobs = t_x_obs.to(self._accelerator), t_ts_obs.to(self._accelerator), t_ts_unobs.to(
            self._accelerator)
        x_unobs_dummy = torch.zeros(1, 1, 4).to(self._accelerator)
        if use_cache:
            # No need to encode trajectory if it is already cached
            t_obs_last = torch.tensor(t_obs_last, dtype=self._dtype)
            t_cache = torch.tensor([t_cache], dtype=self._dtype).view(1, 1)
            z_cache, t_cache = z_cache.to(self._accelerator), t_cache.to(self._accelerator)
            outputs = self._model.filter(z_cache, t_cache, t_obs_last, x_unobs_dummy, t_ts_unobs, mask=False)
            z_cache, t_cache = z_cache.detach().cpu(), t_cache.detach().cpu()
            outputs['z_traj'] = z_cache
        else:
            outputs = self._model.forward(t_x_obs, t_ts_obs, x_unobs_dummy, t_ts_unobs, mask=False)
        outputs = {k: v.detach().cpu() for k, v in outputs.items()}
        x_unobs_dummy.detach().cpu()

        t_x_prior_mean = outputs['prior_mean']
        t_x_prior_var = outputs['prior_logvar']
        z_prior = outputs['z_prior']
        z_traj = outputs['z_traj']

        _, prior_mean, *_ = self._transform.inverse(data=[x_obs, t_x_prior_mean, None], shallow=False)

        prior_mean = prior_mean[-1]  # Removing temporal dimension after inverse transform
        prior_var = self._transform.inverse_var(t_x_prior_var, additional_data=[x_obs, None], shallow=False)
        prior_std = torch.sqrt(prior_var)[-1]

        return buffer, prior_mean, prior_std, z_prior, (buffer.last_obs_time, z_traj)

    def predict(self, state: State) -> State:
        buffer, prior_mean, prior_std, z1_prior, cache = self.multistep_predict(state, n_steps=1)
        return buffer, prior_mean[0], prior_std[0], z1_prior, cache

    def singlestep_to_multistep_state(self, state: State) -> State:
        buffer, prior_mean, prior_std, z1_prior = state
        return buffer, prior_mean.unsqueeze(0), prior_std.unsqueeze(0), z1_prior

    @torch.no_grad()
    def update(self, state: State, measurement: torch.Tensor) -> State:
        buffer, _, _, z_prior, cache = state
        buffer: ObservationBuffer

        x_obs, ts_obs, ts_unobs = buffer.get_input(1)
        buffer.push(measurement)

        if ts_obs.shape[0] == 1:
            # Only one bbox in history - using baseline (propagate last bbox) instead of NN model
            x_unobs_mean_hat, x_unobs_std_hat, z1_prior = self._baseline(x_obs, ts_obs, ts_unobs)
            return buffer, x_unobs_mean_hat, x_unobs_std_hat, z1_prior, cache

        _, t_measurement, *_ = self._transform.apply(data=[x_obs, measurement.view(1, 1, -1), ts_obs, ts_unobs, None], shallow=False)
        t_measurement, z_prior = t_measurement[0].to(self._accelerator), z_prior.to(self._accelerator)
        z_evidence = self._model.encode_unobs(t_measurement)
        t_posterior_mean, t_posterior_log_var, z_posterior = self._model.estimate_posterior(z_prior, z_evidence)
        t_posterior_mean, t_posterior_log_var, z_posterior = \
            t_posterior_mean.detach().cpu(), t_posterior_log_var.detach().cpu(), z_posterior.detach().cpu()
        t_posterior_var = self._model.postprocess_log_var(t_posterior_log_var)
        _, posterior_mean, *_ = self._transform.inverse(data=[x_obs, t_posterior_mean.view(1, 1, -1), None], shallow=False)
        posterior_var = self._transform.inverse_var(t_posterior_var, additional_data=[posterior_mean, None], shallow=False)
        posterior_std = torch.sqrt(posterior_var)

        posterior_mean = posterior_mean[-1, 0, :]
        posterior_std = posterior_std[-1, :]

        return buffer, posterior_mean, posterior_std, z_posterior, cache

    def missing(self, state: State) -> State:
        buffer, mean, std, z_state, cache = state
        buffer.increment()
        return buffer, mean, std, z_state, cache

    def project(self, state: State) -> Tuple[torch.Tensor, torch.Tensor]:
        _, mean, std, _, _ = state
        var = torch.square(std)
        return mean, var

    def _baseline(
        self,
        x_obs: torch.Tensor,
        ts_obs: torch.Tensor,
        ts_unobs: torch.Tensor,
        multistep: bool = False
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        """
        In case there are not enough measurements (e.g. only one measurement is present),
        `baseline` is used instead of a NN model. Baseline just copies last observation.

        Args:
            x_obs: Observations
            ts_obs: Observations' time points
            ts_unobs: Unobserved (prediction) time points
            multistep: Multistep

        Returns:
            - Mean prediction
            - Standard deviation prediction
            - latent state
        """
        x_unobs_mean_hat = torch.stack([x_obs[-1].clone() for _ in range(ts_unobs.shape[0])]).to(ts_obs)
        x_unobs_std_hat = 10 * torch.ones_like(x_unobs_mean_hat).to(ts_obs)

        z0, _ = self._model.initialize(batch_size=1, device=self._accelerator)
        t_obs, t_unobs = self._get_ts()
        t_obs, t_unobs = t_obs.to(self._accelerator), t_unobs.to(self._accelerator)
        _, _, z1_prior = self._model.estimate_prior(z0, t_obs, t_unobs)
        z1_prior = z1_prior.detach().cpu()

        x_unobs_mean_hat, x_unobs_std_hat = x_unobs_mean_hat[:, 0, :], x_unobs_std_hat[:, 0, :]  # Remove batch
        if not multistep:
            x_unobs_mean_hat, x_unobs_std_hat = x_unobs_mean_hat[0], x_unobs_std_hat[0]

        return x_unobs_mean_hat, x_unobs_std_hat, z1_prior


def run_test() -> None:
    # pylint: disable=import-outside-toplevel
    from motrack_motion.datasets.transforms.base import IdentityTransform
    from motrack_motion.models.architectures.rnn_filter import RNNFilterModel

    smf = BufferedE2EFilter(
        model=RNNFilterModel(
            observable_dim=4,
            latent_dim=4,
            output_dim=4,
            t_scale=5.0
        ),
        buffer_size=5,
        accelerator='cpu',
        transform=IdentityTransform()
    )

    measurements = torch.randn(5, 4, dtype=torch.float32)

    # Initiate test
    state = smf.initiate(measurements[0])

    for i in range(1, 5):
        # Predict test
        state = smf.predict(state)
        mean_hat, cov_hat = smf.project(state)
        assert mean_hat.shape == (4,) and cov_hat.shape == (4,)

        # Update test
        state = smf.update(state, measurements[i])
        mean_updated, cov_updated = smf.project(state)
        assert mean_updated.shape == (4,) and cov_updated.shape == (4,)

    for i in range(5, 10):
        # Predict test
        state = smf.predict(state)
        mean_hat, cov_hat = smf.project(state)
        assert mean_hat.shape == (4,) and cov_hat.shape == (4,)

        # Missing test
        state = smf.missing(state)
        mean_updated, cov_updated = smf.project(state)
        assert mean_updated.shape == (4,) and cov_updated.shape == (4,)


if __name__ == '__main__':
    run_test()



"""
Implementation of SOT (single-object-tracking) metrics.
"""
from typing import List, Union, Dict

import numpy as np

ThresholdType = Union[float, List[float], np.ndarray]


def traj_bbox_xywh_to_xyxy(bbox: np.ndarray, eps=1e-9) -> np.ndarray:
    """
    Converts bbox format xywh to xyxy.

    Args:
        bbox: BBox in xywh format.
        eps: minimum width and height

    Returns:
        Bbox in xyxy format
    """
    bbox = bbox.copy()
    bbox[..., 2] = np.maximum(bbox[..., 0] + bbox[..., 2], bbox[..., 0] + eps)
    bbox[..., 3] = np.maximum(bbox[..., 1] + bbox[..., 3], bbox[..., 1] + eps)
    return bbox


def point_to_point_distance(lhs: np.ndarray, rhs: np.ndarray) -> np.ndarray:
    """
    Calculates point to point distance.

    Args:
        lhs: Left hand side point
        rhs: Right hand side point

    Returns:
        Distance (Array)
    """
    return np.sqrt((lhs[..., 0] - rhs[..., 0]) ** 2 + (lhs[..., 1] - rhs[..., 1]) ** 2)


def mse(gt_traj: np.ndarray, pred_traj: np.ndarray) -> float:
    """
    Mean Squared Error.

    Args:
        gt_traj: Ground Truth Trajectory
        pred_traj: Prediction Trajectory

    Returns:
        MSE metric.
    """
    return float(((gt_traj - pred_traj) ** 2).mean())


def iou(bboxes1: np.ndarray, bboxes2: np.ndarray, is_xywh_format: bool = True) -> np.ndarray:
    """
    Calculates iou between each bbox (for each time point and each batch).
    Used as helper function to calculate other IOU based metrics.

    Args:
        bboxes1: lhs bboxes
        bboxes2: rhs bboxes
        is_xywh_format: If true then it is required to be converted to xyxy first

    Returns:
        iou score for each bbox
    """
    if bboxes1.shape != bboxes2.shape:
        raise AttributeError(f'BBoxes do not have the same shape: {bboxes1.shape} != {bboxes2.shape}')

    if bboxes1.size == 0:
        return np.array([], dtype=np.float32)

    if is_xywh_format:
        bboxes1 = traj_bbox_xywh_to_xyxy(bboxes1, eps=0.0)
        bboxes2 = traj_bbox_xywh_to_xyxy(bboxes2, eps=0.0)

    bboxes1_area = (bboxes1[..., 2] - bboxes1[..., 0]) * (bboxes1[..., 3] - bboxes1[..., 1])
    bboxes2_area = (bboxes2[..., 2] - bboxes2[..., 0]) * (bboxes2[..., 3] - bboxes2[..., 1])

    # Calculate the coordinates of the intersection rectangles
    left = np.maximum(bboxes1[..., 0], bboxes2[..., 0])
    up = np.maximum(bboxes1[..., 1], bboxes2[..., 1])
    right = np.minimum(bboxes1[..., 2], bboxes2[..., 2])
    down = np.minimum(bboxes1[..., 3], bboxes2[..., 3])

    # Calculate the area of intersection rectangles
    width = np.maximum(right - left, 0)
    height = np.maximum(down - up, 0)
    intersection_area = width * height

    # Calculate the IoU
    union_area = bboxes1_area + bboxes2_area - intersection_area
    iou_scores = np.divide(intersection_area, union_area, out=np.zeros_like(union_area), where=union_area != 0)

    return iou_scores


def accuracy(gt_traj: np.ndarray, pred_traj: np.ndarray) -> float:
    """
    Calculates average IOU between GT and PRED bbox.

    Args:
        gt_traj: Ground Truth Trajectory
        pred_traj: Prediction Trajectory

    Returns:
        Accuracy metric.
    """
    return iou(gt_traj, pred_traj).mean()


def metrics_func(gt_traj: np.ndarray, pred_traj: np.ndarray) -> Dict[str, float]:
    """
    Calculates Accuracy, Success and NormPrecision. Supports only default metric parameters.

    Args:
        gt_traj: Ground Truth Trajectory
        pred_traj: Prediction Trajectory

    Returns:
        Mappings for each metric
    """
    return {
        'Accuracy': accuracy(gt_traj, pred_traj),
        'MSE': mse(gt_traj, pred_traj)
    }



"""
Set of neural network building blocks
"""
from motrack_motion.library.building_blocks.mlp import MLP
from motrack_motion.library.building_blocks.attention import TemporalFirstMultiHeadSelfAttention, TemporalFirstMultiHeadCrossAttention



"""
Dataset, model and inference file system structure

Pipeline data structure:
{master_path}/
    {dataset_name}/
        {experiment_name}/
            checkpoints/
                {checkpoint_name_1}
                {checkpoint_name_2}
                ...
            run_history/*
            configs/
                [train.yaml]
                [inference.yaml] (last inference config)
                [visualize.yaml]
                ...
        tensorboard_logs/
            {experiment_name}/*
    analysis/*
"""
import os.path

CHECKPOINTS_DIRNAME = 'checkpoints'
RUN_HISTORY_DIRNAME = 'run_history'
INFERENCES_DIRNAME = 'inferences'
INFERENCE_VISUALIZATIONS_DIRNAME = 'visualizations'
TENSORBOARD_DIRNAME = 'tensorboard_logs'
CONFIGS_DIRNAME = 'configs'
ANALYSIS_DIRNAME = 'analysis'


def get_experiment_path(master_path: str, dataset_name: str, experiment_name: str) -> str:
    """
    Gets experiment location given 'master_path'

    Args:
        master_path: Global path (master path)
        dataset_name: Dataset name
        experiment_name: Model experiment name

    Returns:
        Experiment absolute path
    """
    return os.path.join(master_path, dataset_name, experiment_name)


def get_run_history_path(experiment_path) -> str:
    """
    Gets experiment's run history.

    Args:
        experiment_path: Experiment path

    Returns:
        Experiment's run history
    """
    return os.path.join(experiment_path, RUN_HISTORY_DIRNAME)


def get_checkpoints_dirpath(experiment_path: str) -> str:
    """
    Gets path to all checkpoints for given experiment

    Args:
        experiment_path: Path to experiment

    Returns:
        Checkpoints directory
    """
    return os.path.join(experiment_path, CHECKPOINTS_DIRNAME)


def get_checkpoint_path(experiment_path: str, checkpoint_filename: str) -> str:
    """
    Gets checkpoint full path.

    Args:
        experiment_path: Path to experiment
        checkpoint_filename: Checkpoint name

    Returns:
        Checkpoint full path
    """
    return os.path.join(get_checkpoints_dirpath(experiment_path), checkpoint_filename)


def get_config_dirpath(experiment_path: str) -> str:
    """
    Gets path to experiment configs.

    Args:
        experiment_path: Experiment path

    Returns:
        Path to saved configs (history) for given experiment path
    """
    return os.path.join(experiment_path, CONFIGS_DIRNAME)


def get_config_path(experiment_path: str, config_name: str) -> str:
    """
    Gets configs full path.

    Args:
        experiment_path: Path to experiment
        config_name: Config name

    Returns:
        Config full path
    """
    return os.path.join(get_config_dirpath(experiment_path), config_name)


def get_tensorboard_logs_dirpath(master_path: str, dataset_name: str) -> str:
    """
    Gets location where all tensorboard logs for one dataset are stored.

    Args:
        master_path: Master path
        dataset_name: Dataset name

    Returns:
        Tensorboard directory path
    """
    return os.path.join(master_path, dataset_name, TENSORBOARD_DIRNAME)



"""
Config structure. Config should be loaded as dictionary and parsed into GlobalConfig Python object. Benefits:
- Structure and type validation (using dacite library)
- Custom validations
- Python IDE autocomplete
"""
import logging
import os
from dataclasses import field
from typing import Optional, Dict, Any

from hydra.core.config_store import ConfigStore
from hydra.utils import instantiate
from omegaconf import OmegaConf
from pydantic.dataclasses import dataclass

from motrack_motion.common import project, conventions
from motrack_motion.datasets.augmentations import create_identity_augmentation_config

logger = logging.getLogger('ConfigParser')


@dataclass
class DatasetConfig:
    """
    Dataset config.
    - name: Dataset name
    - path: Path to the dataset
    - history_len: Number of observable input values
    - future_len: Number of unobservable output values (values that are being predicted)
    """
    name: str
    path: str
    history_len: int
    future_len: int

    additional_params: Optional[Dict[str, Any]] = field(default_factory=dict)

    def __post_init__(self) -> None:
        """
        Postprocess and validation.
        """
        self.fullpath = None

    def get_dataset_split_path(self, split: str) -> str:
        """
        Gets dataset split path with MOT conventions.

        Args:
            split: Split name

        Returns:
            Dataset split path
        """
        assert split in ['train', 'val', 'test'], f'Invalid split name: {split}!'
        return os.path.join(self.fullpath, split)


@dataclass
class TransformConfig:
    """
    Features transform config. Used to transform input features. Optionally performs inverse operation to output values.
    - name: Name of transform class (type).
    - params: Transform parameters
    """
    name: str
    params: dict


@dataclass
class AugmentationsConfig:
    """
    Augmentations applied before or after transform:
    - before_transform: Composition of augmentations applied before transform function.
    - after_transform: Composition of augmentations applied after transform function.
    """
    before_transform_config: Optional[dict]
    after_transform_config: Optional[dict]
    after_batch_collate_config: Optional[dict]

    def __post_init__(self):
        """
        Validation: Check augmentation object instantiation
        """
        self.before_transform_config, self.after_transform_config, self.after_batch_collate_config = \
            [create_identity_augmentation_config() if cfg is None else cfg
             for cfg in [self.before_transform_config, self.after_transform_config, self.after_batch_collate_config]]

        self.before_transform = instantiate(OmegaConf.create(self.before_transform_config))
        self.after_transform = instantiate(OmegaConf.create(self.after_transform_config))
        self.after_batch_collate = instantiate(OmegaConf.create(self.after_batch_collate_config))

    @classmethod
    def default(cls) -> 'AugmentationsConfig':
        """
        Default augmentations (none) in case it is not defined.

        Returns:
            Default augmentations
        """
        # noinspection PyArgumentList
        return cls(
            before_transform_config=create_identity_augmentation_config(),
            after_transform_config=create_identity_augmentation_config(),
            after_batch_collate_config=create_identity_augmentation_config()
        )


@dataclass
class ModelConfig:
    """
    Model config:
    - type: Model (architecture) type
    - params: Model creation parameters
    """
    type: str
    params: dict


@dataclass
class ResourcesConfig:
    """
    Resources config (cpu/gpu, number of cpu cores, ...)
    - gpus: Number of gpus
    - accelerator: gpu/cpu
    - num_workers: cpu workers
    """
    devices: int
    accelerator: str
    num_workers: int


@dataclass
class TrainLoggingConfig:
    """
    Configs for script logging during model training (not important for inference).
    - path: TB logs path (deprecated)
    - log_every_n_steps: TB log frequency
    """
    path: str  # Deprecated (predefined by conventions)
    log_every_n_steps: int


@dataclass
class TrainCheckpointConfig:
    """
    Model checkpoint saving config.
    - metric_monitor: Chooses the best checkpoint based on metric name
    - resume_from: Start from chosen checkpoint (fine-tuning)
    """
    metric_monitor: str
    resume_from: Optional[str]


@dataclass
class TrainConfig:
    """
    Train configuration.
    - experiment: Name of the training experiment
    - batch_size: Training batch size
    - max_epochs: Number of epochs to train model

    - logging_cfg: TrainLoggingConfig
    - checkpoint_cfg: TrainCheckpointConfig

    - train_params: Training architecture specific parameters
    """
    type: str
    params: dict

    batch_size: int
    max_epochs: int

    logging_cfg: TrainLoggingConfig
    checkpoint_cfg: TrainCheckpointConfig

    inverse_transform_before_loss: bool = field(default=False)
    train_params: Optional[dict] = field(default_factory=dict)  # Model train params

    gradient_clip_val: Optional[float] = field(default=None)
    gradient_clip_algorithm: Optional[str] = field(default=None)

    train_on_val: bool = field(default=False)


@dataclass
class EvalConfig:
    """
    Inference + Evaluation config
    - batch_size: inference batch size
    - inference_name: inference run name
    - split: what split to use for evaluation (train/val/test)
    - checkpoint: what checkpoint to use for evaluation

    - autoregressive: use autoregressive decorator
    - autoregressive_keep_history: requires `autoregressive` - keeps all history when predicting (not dropping last)
    """
    batch_size: int
    inference_name: str
    split: str
    checkpoint: Optional[str]

    # Autoregressive configs
    autoregressive: bool = field(default=False)
    autoregressive_keep_history: bool = field(default=False)

    # Override dataset configs
    dataset_future_len: Optional[int] = field(default=None)

    # Custom evaluation
    fps_multiplier: float = field(default=1.0)


@dataclass
class PathConfig:
    """
    Path configs
    - master: location where all final and intermediate results are stored
    - assets: location where datasets can be found
    """
    master: str
    assets: str

    @classmethod
    def default(cls) -> 'PathConfig':
        """
        Default path configuration is used if it is not defined in configs.

        Returns: Path configuration.
        """
        # noinspection PyArgumentList
        return cls(
            master=project.OUTPUTS_PATH,
            assets=project.ASSETS_PATH
        )


@dataclass
class GlobalConfig:
    """
    Scripts GlobalConfig
    """
    experiment: str
    resources: ResourcesConfig
    dataset: DatasetConfig
    transform: TransformConfig
    train: TrainConfig
    eval: EvalConfig
    model: ModelConfig

    augmentations: AugmentationsConfig = field(default_factory=AugmentationsConfig.default)
    path: PathConfig = field(default_factory=PathConfig.default)

    def __post_init__(self) -> None:
        """
        Postprocessing.
        """
        self.dataset.fullpath = os.path.join(self.path.assets, self.dataset.path)

    @property
    def experiment_path(self) -> str:
        """
        Get experiment path.

        Returns:
            Experiments path
        """
        return conventions.get_experiment_path(self.path.master, self.dataset.path, self.experiment)

# Configuring hydra config store
# If config has `- global_config` in defaults then
# full config is recursively instantiated
cs = ConfigStore.instance()
cs.store(name='global_config', node=GlobalConfig)



"""
End-to-end filter training module.
"""
from typing import Optional, Union, Tuple, Dict

import torch

from motrack_motion.datasets.transforms import InvertibleTransform, InvertibleTransformWithVariance
from motrack_motion.models.architectures.transfilter import TransFilter, pad_sequence
from motrack_motion.models.losses import factory_loss_function
from motrack_motion.models.train.base import LightningModuleBase, LightningTrainConfig
from motrack_motion.models.train.metrics import metrics_func


class TransFilterTrainingModule(LightningModuleBase):
    """
    PytorchLightning wrapper for TransFilter model.
    """
    def __init__(
        self,
        model: TransFilter,

        transform_func: Optional[Union[InvertibleTransform, InvertibleTransformWithVariance]] = None,

        prior_loss_weight: float = 1.0,
        posterior_loss_weight: float = 1.0,

        train_config: Optional[LightningTrainConfig] = None
    ):
        super().__init__(train_config=train_config)

        assert isinstance(model, TransFilter), f'Expected FilterModule but found {type(model)}'
        self._model = model

        # Loss weights
        self._prior_loss_weight = prior_loss_weight
        self._posterior_loss_weight = posterior_loss_weight

        self._loss_func = factory_loss_function(train_config.loss_name, train_config.loss_params) \
            if train_config is not None else None
        if transform_func is not None:
            assert isinstance(transform_func, InvertibleTransformWithVariance), \
                f'Expected transform function to be of type "InvertibleTransformWithStd" ' \
                f'but got "{type(transform_func)}"'
        self._transform_func = transform_func

    @property
    def core(self) -> TransFilter:
        """
        Get core model (without the training module).

        Returns:
            Core model
        """
        return self._model

    def forward(self, x_obs: torch.Tensor, t_obs: torch.Tensor, x_unobs: torch.Tensor, t_unobs: torch.Tensor, mask: Optional[torch.Tensor] = None) \
            -> Dict[str, torch.Tensor]:
        return self._model(x_obs, t_obs, x_unobs, t_unobs, mask=mask)

    def inference(self, x_obs: torch.Tensor, t_obs: torch.Tensor, x_unobs: torch.Tensor, t_unobs: torch.Tensor, mask: Optional[torch.Tensor] = None) \
            -> Dict[str, torch.Tensor]:
        """
        Inference (alias for forward)

        Args:
            x_obs: Observed data
            t_obs: Observed time points
            x_unobs: Unobserved data
            t_unobs: Unobserved time points
            mask: Transformer mask

        Returns:
            Prior and posterior estimation for future trajectory
        """
        return self._model(x_obs, t_obs, x_unobs, t_unobs, mask=mask)

    def _calc_loss_and_metrics(
        self,
        orig_bboxes_obs: torch.Tensor,
        orig_bboxes_unobs_prior: torch.Tensor,
        orig_bboxes_unobs_posterior: torch.Tensor,
        transformed_bboxes_unobs: torch.Tensor,
        bboxes_unobs_prior_mean: torch.Tensor,
        bboxes_unobs_posterior_mean: torch.Tensor,
        metadata: dict
    ) -> Tuple[Union[torch.Tensor, Dict[str, torch.Tensor]], Optional[Dict[str, float]]]:
        prior_loss = self._loss_func(bboxes_unobs_prior_mean, transformed_bboxes_unobs)
        posterior_loss = self._loss_func(bboxes_unobs_posterior_mean, transformed_bboxes_unobs)

        if isinstance(prior_loss, dict) and isinstance(posterior_loss, dict):
            prior_loss, posterior_loss = prior_loss['loss'], posterior_loss['loss']
        loss = {
            'prior_loss': prior_loss,
            'posterior_loss': posterior_loss,
            'loss': self._prior_loss_weight * prior_loss + self._posterior_loss_weight * posterior_loss
        }

        if self._transform_func is not None:
            # Invert mean
            _, bboxes_unobs_prior_mean, *_ = self._transform_func.inverse([orig_bboxes_obs, bboxes_unobs_prior_mean, metadata, None],
                                                                          shallow=False)
            _, bboxes_unobs_posterior_mean, *_ = self._transform_func.inverse([orig_bboxes_obs, bboxes_unobs_posterior_mean, metadata, None],
                                                                              shallow=False)

        gt_traj = orig_bboxes_unobs_prior.detach().cpu().numpy()
        prior_traj = bboxes_unobs_prior_mean.detach().cpu().numpy()
        prior_metrics = metrics_func(gt_traj, prior_traj)
        prior_metrics = {f'prior_{name}': value for name, value in prior_metrics.items()}

        gt_traj = orig_bboxes_unobs_posterior.detach().cpu().numpy()
        posterior_traj = bboxes_unobs_posterior_mean.detach().cpu().numpy()
        posterior_metrics = metrics_func(gt_traj, posterior_traj)
        posterior_metrics = {f'posterior_{name}': value for name, value in posterior_metrics.items()}
        metrics = dict(list(prior_metrics.items()) + list(posterior_metrics.items()))

        return loss, metrics

    def _step(self, batch: Dict[str, Union[dict, torch.Tensor]], prefix: str) -> torch.Tensor:
        bboxes_obs, bboxes_aug_unobs, ts_obs, ts_unobs, orig_bboxes_obs, orig_bboxes_unobs, bboxes_unobs, metadata = batch.values()

        with torch.no_grad():
            bboxes_obs, ts_obs, mask = pad_sequence(bboxes_obs, ts_obs)

        outputs = self.forward(bboxes_obs, ts_obs, bboxes_aug_unobs, ts_unobs, mask=mask)

        loss, metrics = self._calc_loss_and_metrics(
            orig_bboxes_obs=orig_bboxes_obs,
            orig_bboxes_unobs_prior=orig_bboxes_unobs,
            orig_bboxes_unobs_posterior=orig_bboxes_unobs,
            transformed_bboxes_unobs=bboxes_unobs,
            bboxes_unobs_prior_mean=outputs['x_prior'],
            bboxes_unobs_posterior_mean=outputs['x_posterior'],
            metadata=metadata
        )
        self._log_loss(loss, prefix=prefix, log_step=True)
        self._log_metrics(metrics, prefix=prefix)

        return loss

    def training_step(self, batch: Dict[str, Union[dict, torch.Tensor]], *args, **kwargs) -> torch.Tensor:
        loss = self._step(batch, prefix='training')
        self._log_lr()
        return loss

    def validation_step(self, batch: Dict[str, Union[dict, torch.Tensor]], *args, **kwargs) -> torch.Tensor:
        loss = self._step(batch, prefix='val')
        return loss



"""
Diff stats analysis
"""
import logging
from collections import defaultdict

import hydra
import torch
import yaml
from tqdm import tqdm

from motrack_motion.common.project import CONFIGS_PATH
from motrack_motion.config_parser import GlobalConfig
from motrack_motion.datasets import TorchTrajectoryDataset, dataset_factory
from motrack_motion.datasets import transforms
from motrack_motion.utils import pipeline

logger = logging.getLogger('DatasetDiffStats')


@hydra.main(config_path=CONFIGS_PATH, config_name='default', version_base='1.1')
@pipeline.task('dataset-stats')
def main(cfg: GlobalConfig):

    dataset = TorchTrajectoryDataset(
        dataset_factory(
            name=cfg.dataset.name,
            paths=cfg.dataset.get_dataset_split_path('train'),
            history_len=cfg.dataset.history_len,
            future_len=cfg.dataset.future_len,
            additional_params=cfg.dataset.additional_params
        ),
        transform=transforms.BBoxRelativeToLastObsTransform(scale_by_time_diff=True)  # Choose your transformation
    )

    category_data = defaultdict(lambda: {
        'sum_bbox': torch.zeros(4, dtype=torch.float32),
        'sum_bbox2': torch.zeros(4, dtype=torch.float32),
        'n_total': 0
    })

    sum_bbox = torch.zeros(4, dtype=torch.float32)  # Used to calculate mean and std
    sum_bbox2 = torch.zeros(4, dtype=torch.float32)  # Used to calculate std
    n_total = 0

    # noinspection PyTypeChecker
    for sample in tqdm(dataset, unit='sample', desc='Calculating diff statistics'):
        bboxes_obs, bboxes_unobs, _, _, _, _, _, metadata = sample.values()
        category = metadata['category']

        for bboxes in [bboxes_obs, bboxes_unobs]:
            # General stats
            sum_bbox += bboxes.sum(dim=0)
            sum_bbox2 += torch.pow(bboxes, 2).sum(dim=0)
            n_total += bboxes.shape[0]

            # Category stats
            category_data[category]['sum_bbox'] += bboxes.sum(dim=0)
            category_data[category]['sum_bbox2'] += torch.pow(bboxes, 2).sum(dim=0)
            category_data[category]['n_total'] += bboxes.shape[0]

    # Calculate statistics (global)
    mean_bbox = sum_bbox / n_total
    mean_bbox2 = sum_bbox2 / n_total
    std_bbox = (mean_bbox2 - mean_bbox ** 2) ** 0.5

    # Convert all to lists
    mean_bbox, mean_bbox2, std_bbox = [v.numpy().tolist() for v in [mean_bbox, mean_bbox2, std_bbox]]

    # Calculate statistics (category)
    for cat_data in category_data.values():
        cat_data['mean'] = cat_data['sum_bbox'] / cat_data['n_total']
        cat_data['mean2'] = cat_data['sum_bbox2'] / cat_data['n_total']
        cat_data['std'] = (cat_data['mean2'] - cat_data['mean'] ** 2) ** 0.5

        del cat_data['n_total']
        del cat_data['sum_bbox']
        del cat_data['sum_bbox2']
        del cat_data['mean2']

        cat_data['mean'] = cat_data['mean'].numpy().tolist()
        cat_data['std'] = cat_data['std'].numpy().tolist()


    stats_data = {
        'mean': mean_bbox,
        'std': std_bbox,
        'by_category': dict(category_data)
    }

    logger.info(f'Stats\n{yaml.dump(stats_data)}')


if __name__ == '__main__':
    main()



"""
Motrack motion filter interface.
"""
from motrack_motion.filter.end_to_end_filter import BufferedE2EFilter
from motrack_motion.filter.transfilter import TransFilter
from motrack_motion.filter.factory import filter_factory



"""
Implementation of multi-head attention from "Attention is all you need" paper
"""
import math
from typing import Optional

import torch
from torch import nn


class DotProductAttention(nn.Module):
    """
    Dot-scaled product attention.
    """
    NEG_INF = -999_999

    def forward(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor, mask: Optional[torch.Tensor] = None) -> torch.Tensor:
        _, _, _, dim = k.shape

        k_t = torch.transpose(k, 2, 3)
        scaled_dot_product = (q @ k_t) / math.sqrt(dim)  # q @ k.T / sqrt(dim)

        if mask is not None:
            scaled_dot_product = torch.masked_fill(scaled_dot_product, mask=mask, value=self.NEG_INF)

        score = torch.softmax(scaled_dot_product, dim=-1)  # softmax(q @ k.T / sqrt(dim))
        return score @ v  # softmax(q @ k.T / sqrt(dim)) @ v


class MultiHeadAttention(nn.Module):
    """
    Implementation of multi-head attention without projections.
    Base for SelfAttention and CrossAttention.
    Ref: https://arxiv.org/pdf/1706.03762.pdf
    """
    def __init__(self, n_heads: int):
        super().__init__()
        self._n_heads = n_heads
        self._attention = DotProductAttention()

    def _multi_head_attention(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
        """
        Multi head attention without projections.

        Args:
            q: Query
            k: Keys
            v: Values

        Returns:
            Attention result
        """
        q, k, v = self._split(q), self._split(k), self._split(v)
        o = self._attention(q, k, v)
        o = self._concat(o)
        return o

    def _split(self, x: torch.Tensor) -> torch.Tensor:
        """
        Splits input tensor dimension by number of head (over batch and temporal dimension)

        Args:
            x: Input tensor

        Returns:
            Output tensor
        """
        batch_size, time_len, full_dim = x.shape  # B x T x (H * D)
        assert full_dim % self._n_heads == 0, \
            f'Dimension ({full_dim}) is not divisible by number of heads ({self._n_heads}).'

        head_dim = full_dim // self._n_heads
        x = x.reshape(batch_size, time_len, self._n_heads, head_dim)    # B x T x H x D
        x = torch.transpose(x, 1, 2).contiguous()  # B x H x T x D
        return x

    def _concat(self, x: torch.Tensor) -> torch.Tensor:
        """
        Concatenates input tensor that was previously split.

        Args:
            x: Split tensor

        Returns:
            Combined tensor
        """
        batch_size, n_heads, time_len, head_dim = x.shape  # B x H x T x D
        assert n_heads == self._n_heads

        x = torch.transpose(x, 1, 2)  # B x T x H x D
        x = x.reshape(batch_size, time_len, n_heads * head_dim) # B x T x (H * D)
        return x


class MultiHeadSelfAttention(MultiHeadAttention):
    """
    Implementation of self multi-head attention.
    """
    def __init__(self, n_heads: int, dim: int):
        """
        Args:
            n_heads: Number of attention heads
            dim: Input/output dimension
        """
        super().__init__(n_heads=n_heads)
        self._n_heads = n_heads
        self._w_q = nn.Linear(dim, dim)
        self._w_k = nn.Linear(dim, dim)
        self._w_v = nn.Linear(dim, dim)
        self._w_o = nn.Linear(dim, dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        q, k, v = self._w_q(x), self._w_k(x), self._w_v(x)  # project input
        o = self._multi_head_attention(q, k, v)
        o = self._w_o(o)  # project output
        return o


class MultiHeadCrossAttention(MultiHeadAttention):
    """
    Implementation of cross multi-head attention.
    """
    def forward(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
        return self._multi_head_attention(q, k, v)


class TemporalFirstMultiHeadSelfAttention(nn.Module):
    """
    Wrapper for `MultiHeadSelfAttention` in case of "batch_first=False" input.
    """
    def __init__(self, n_heads: int, dim: int):
        """
        Args:
            n_heads: Number of attention heads
            dim: Input/output dimension
        """
        super().__init__()
        self._mhsa = MultiHeadSelfAttention(n_heads=n_heads, dim=dim)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        x = torch.transpose(x, 0, 1).contiguous()
        x = self._mhsa(x)
        x = torch.transpose(x, 0, 1).contiguous()
        return x


class TemporalFirstMultiHeadCrossAttention(nn.Module):
    """
    Wrapper for `MultiHeadCrossAttention` in case of "batch_first=False" input.
    """
    def __init__(self, n_heads: int):
        """
        Args:
            n_heads: Number of attention heads
        """
        super().__init__()
        self._mhca = MultiHeadCrossAttention(n_heads=n_heads)

    def forward(self, q: torch.Tensor, k: torch.Tensor, v: torch.Tensor) -> torch.Tensor:
        q, k, v = [torch.transpose(value, 0, 1).contiguous() for value in [q, k, v]]
        x = self._mhca(q, k, v)
        x = torch.transpose(x, 0, 1)
        return x


def run_test() -> None:
    mha = MultiHeadSelfAttention(4, 16)
    x = torch.randn(10, 3, 16)
    o = mha(x)
    print(f'Output shape: {o.shape}')

    tf_mhsa = TemporalFirstMultiHeadSelfAttention(4, 16)
    x = torch.randn(3, 10, 16)
    o = tf_mhsa(x)
    print(f'Output shape (: {o.shape}')


if __name__ == '__main__':
    run_test()



"""
Implementation of the RNNFilter.
"""
from typing import Tuple, Union

import torch
from torch import nn

from motrack_motion.library.building_blocks.mlp import MLP
from motrack_motion.models.architectures.end_to_end import FilterModule


class RNNMultiStepModule(nn.Module):
    """
    RNNMultiStepModule
    """
    def __init__(
        self,
        latent_dim: int,
        n_rnn_layers: int = 1,
    ):
        """
        Args:
            latent_dim: "latent" (hidden-2) trajectory dimension
            n_rnn_layers: Number of stacked RNN (GRU) layers
        """
        super().__init__()

        self._latent_dim = latent_dim
        self._n_rnn_layers = n_rnn_layers

        self._rnn = nn.GRU(
            input_size=1,
            hidden_size=latent_dim,
            num_layers=n_rnn_layers,
            batch_first=False
        )

    def forward(self, z: torch.Tensor, t: torch.Tensor) -> torch.Tensor:
        n_steps = round(float(t[0, 0, 0]))

        z = z.unsqueeze(0)
        for t_i in range(1, n_steps + 1):
            t_i = torch.tensor([t_i], dtype=torch.float32).view(1, 1, 1).repeat(1, z.shape[1], 1).to(z)
            _, z = self._rnn(t_i, z)

        return z[-1]


class RNNFilterModel(FilterModule):
    """
    RNNFilterModel
    """
    def __init__(
        self,
        observable_dim: int,
        latent_dim: int,
        output_dim: int,

        bounded_variance: bool = False,
        bounded_value: float = 0.01,

        n_rnn_layers: int = 1,
        n_update_layers: int = 1,
        n_head_mlp_layers: int = 2,
        n_obs2latent_mlp_layers: int = 1,
        t_scale: float = 30.0
    ):
        super().__init__(
            bounded_variance=bounded_variance,
            bounded_value=bounded_value,
            t_scale=t_scale
        )
        self._bounded_variance = bounded_variance

        self._observable_dim = observable_dim
        self._latent_dim = latent_dim
        self._output_dim = output_dim

        self._rnn = RNNMultiStepModule(
            latent_dim=latent_dim,
            n_rnn_layers=n_rnn_layers
        )

        self._prior_head_mean = nn.Sequential(
            MLP(latent_dim, latent_dim, n_layers=n_head_mlp_layers),
            nn.Linear(latent_dim, output_dim, bias=True)
        )
        self._prior_head_log_var = nn.Sequential(
            MLP(latent_dim, latent_dim, n_layers=n_head_mlp_layers),
            nn.Linear(latent_dim, output_dim, bias=True)
        )
        self._posterior_head_mean = nn.Sequential(
            MLP(latent_dim, latent_dim, n_layers=n_head_mlp_layers),
            nn.Linear(latent_dim, output_dim, bias=True)
        )
        self._posterior_head_log_var = nn.Sequential(
            MLP(latent_dim, latent_dim, n_layers=n_head_mlp_layers),
            nn.Linear(latent_dim, output_dim, bias=True)
        )

        self._obs2latent = MLP(
            input_dim=observable_dim,
            hidden_dim=latent_dim,
            output_dim=latent_dim,
            n_layers=n_obs2latent_mlp_layers
        )

        self._unobs2latent = MLP(
            input_dim=output_dim,
            hidden_dim=latent_dim,
            output_dim=latent_dim,
            n_layers=n_obs2latent_mlp_layers
        )

        self._n_update_layers = n_update_layers
        self._update_layer = nn.GRU(
            input_size=latent_dim,
            hidden_size=latent_dim,
            num_layers=n_update_layers,
            batch_first=False
        )

    def initialize(self, batch_size: int, device: Union[str, torch.device]) -> Tuple[torch.Tensor, torch.Tensor]:
        z0 = torch.zeros(batch_size, self._latent_dim, dtype=torch.float32).to(device)
        t0 = torch.zeros(batch_size, 1, dtype=torch.float32).to(device)
        return z0, t0

    def estimate_prior(self, z0: torch.Tensor, t0: torch.Tensor, t1: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        t = (t1 - t0).unsqueeze(0)
        z1_prior = self._rnn(z0, t)
        x1_prior_mean = self._prior_head_mean(z1_prior)
        x1_prior_log_var = self._prior_head_log_var(z1_prior)
        return x1_prior_mean, x1_prior_log_var, z1_prior

    def estimate_posterior(self, z1_prior: torch.Tensor, z1_evidence: torch.Tensor) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        z1_prior = z1_prior.unsqueeze(0).repeat(self._n_update_layers, 1, 1)
        z1_evidence = z1_evidence.unsqueeze(0)
        z1_posterior, _ = self._update_layer(z1_evidence, z1_prior)
        z1_posterior = z1_posterior[0]

        x1_posterior_mean = self._posterior_head_mean(z1_posterior)
        x1_posterior_log_var = self._posterior_head_log_var(z1_posterior)
        return x1_posterior_mean, x1_posterior_log_var, z1_posterior

    def encode_obs(self, x: torch.Tensor) -> torch.Tensor:
        return self._obs2latent(x)

    def encode_unobs(self, x: torch.Tensor) -> torch.Tensor:
        return self._unobs2latent(x)


def run_test() -> None:
    rfm = RNNFilterModel(
        observable_dim=4,
        latent_dim=3,
        output_dim=4,

        n_rnn_layers=2
    )

    x_obs = torch.randn(4, 3, 4)
    x_unobs = torch.randn(2, 3, 4)
    t_obs = torch.tensor([0, 1, 2, 4], dtype=torch.float32).view(-1, 1, 1).repeat(1, 3, 1)
    t_unobs = torch.tensor([6, 9], dtype=torch.float32).view(-1, 1, 1).repeat(1, 3, 1)

    priors_mean, priors_log_var, posteriors_mean, posteriors_log_var = rfm(x_obs, t_obs, x_unobs, t_unobs)
    print(priors_mean.shape, priors_log_var.shape, posteriors_mean.shape, posteriors_log_var.shape)


if __name__ == '__main__':
    run_test()



"""
Filter's observation buffer.
"""
from abc import ABC
from typing import Tuple, List

import torch
from torch import nn

from motrack_motion.datasets import transforms
from motrack_motion.filter.base import StateModelFilter, State
from motrack_motion.library.numpy_utils.bbox import affine_transform


class ObservationBuffer:
    """
    Buffers trajectory observations.
    """
    def __init__(
        self,
        size: int,
        min_size: int,
        min_history: int = 1,
        dtype: torch.dtype = torch.float32,
    ):
        assert size >= 1, f'Invalid size {size}. Minimum size is 1.'

        self._size = size
        self._min_size = min_size
        self._dtype = dtype
        self._min_history = min_history

        self._buffer: List[Tuple[int, torch.Tensor]] = []
        self._t = 0

    @property
    def last_obs_time(self) -> int:
        return self._buffer[-1][0]

    @property
    def time(self) -> int:
        return self._t

    @property
    def has_input(self) -> bool:
        return len(self._buffer) >= self._min_size

    def push(self, x: torch.Tensor) -> None:
        self._buffer.append((self._t, x))
        while len(self._buffer) > self._min_history and (self._t - self._buffer[0][0]) >= self._size:
            self._buffer.pop(0)

        self.increment()

    def increment(self) -> None:
        self._t += 1

    def get_input(self, n_future_steps: int) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor]:
        n_hist_steps = len(self._buffer)
        ts_obs, x_obs = zip(*self._buffer)
        ts_obs_first = ts_obs[0]

        # Form history trajectory
        x_obs = torch.stack(x_obs).view(n_hist_steps, 1, -1)
        ts_obs = torch.tensor(ts_obs, dtype=self._dtype).view(-1, 1, 1)

        # Form estimation trajectory time interval
        ts_unobs = torch.tensor(list(range(self._t, self._t + n_future_steps)),
                                dtype=self._dtype).view(-1, 1, 1)

        ts_obs = ts_obs - ts_obs_first + 1
        ts_unobs = ts_unobs - ts_obs_first + 1

        return x_obs, ts_obs, ts_unobs

    def clear(self) -> None:
        self._buffer.clear()

    def __len__(self) -> int:
        return len(self._buffer)

    def affine_transform(self, warp: torch.Tensor) -> None:
        warp_numpy = warp.cpu().numpy()
        for i in range(len(self)):
            time, measurement = self._buffer[i]
            measurement_numpy = measurement.cpu().numpy()
            measurement_numpy[2:] = measurement_numpy[:2] + measurement_numpy[2:]  # xywh -> xyxy
            warped_measurement_numpy = affine_transform(warp_numpy, measurement_numpy)
            warped_measurement_numpy[2:] = warped_measurement_numpy[2:] - warped_measurement_numpy[:2]  # xyxy -> xywh
            warped_measurement = torch.from_numpy(warped_measurement_numpy)
            self._buffer[i] = (time, warped_measurement)


class BufferedStateModel(StateModelFilter, ABC):
    """
    Buffered state model
    """
    def __init__(
        self,
        model: nn.Module,
        transform: transforms.InvertibleTransformWithVariance,
        accelerator: str,

        buffer_size: int,
        buffer_min_size: int = 1,
        buffer_min_history: int = 5,
        dtype: torch.dtype = torch.float32,

        recursive_inverse: bool = False,
        uncertainty_multiplier: float = 10.0
    ):
        """
        Args:
            model: Model
            transform: Transform function
            accelerator: cpu/gpu accelerator
            buffer_size: Buffer size
            buffer_min_size: Minimum size to run model (default 1)
            buffer_min_history: Minimum history to save
            dtype: Torch datatype
            recursive_inverse: Apply recursive inverse transform
            uncertainty_multiplier: Uncertainty multiplier in case model does not estimate the uncertainty
        """
        self._model = model
        self._transform = transform
        self._accelerator = accelerator
        self._dtype = dtype

        self._buffer_size = buffer_size
        self._buffer_min_size = buffer_min_size
        self._buffer_min_history = buffer_min_history

        # Model
        self._model.to(self._accelerator)
        self._model.eval()

        self._recursive_inverse = recursive_inverse
        self._uncertainty_multiplier = uncertainty_multiplier

    def _create_buffer(self, measurement: torch.Tensor) -> ObservationBuffer:
        """
        Creates buffer with initial measurement.

        Args:
            measurement: Initial buffer measurement

        Returns:
            Buffer
        """
        buffer = ObservationBuffer(
            size=self._buffer_size,
            min_size=self._buffer_min_size,
            min_history=self._buffer_min_history,
            dtype=self._dtype
        )
        buffer.push(measurement)
        return buffer

    def affine_transform(self, state: State, warp: torch.Tensor) -> State:
        if isinstance(state, ObservationBuffer):
            buffer = state
        elif isinstance(state, (tuple, list)):
            buffer_candidates = [s for s in state if isinstance(s, ObservationBuffer)]
            assert len(buffer_candidates) == 1, 'Buffer not found in the state!'
            buffer = buffer_candidates[0]
        else:
            raise AssertionError('Buffer not found in the state!')

        buffer.affine_transform(warp)

        return state



"""
File System custom support functions.
"""
import os
import re
from typing import List, Optional


def listdir(path: str, regex_filter: Optional[str] = None) -> List[str]:
    """
    Wrapper for `os.listdir` that ignore `.*` files (like `.DS_Store)

    Args:
        path: Directory path
        regex_filter: Filter files by regex

    Returns:
        Listed files (not hidden)
    """
    files = [p for p in os.listdir(path) if not p.startswith('.')]
    if regex_filter is not None:
        files = [p for p in os.listdir(path) if re.match(regex_filter, p)]
    return files
